<img src="logo.png">

Parallel Radix Sort - CUDA
&middot;
[![GitLab license](https://img.shields.io/github/license/Day8/re-frame.svg)](LICENSE)
=====
> A non-comparative integer sorting algorithm that sorts data with integer keys by grouping keys by the individual digits which share the same significant position and value.

## Table Of Contents
- [Introduction](#introduction)
- [Requirements](#requirements)
- [Run Radix Sort](#run-radix-sort)
- [Details of Radix Sort (Indonesian)](#details-of-radix-sort-(indonesian))
- [Author](#author)
- [Words From Author](#words-from-author)

## Introduction
This is a simple implementation of parallel radix sort algorithm using CUDA.

## Requirements
* CUDA

## Run Radix Sort
1. Compile Radix Sort Program with `make`.
2. Run `./serial_radix_sort <array_length>` for serial radix sort.
3. Run `./radix_sort <array_length>` for parallel radix sort.
4. You will see the performance for each radix sort implementations.

## Details of Radix Sort (Indonesian)
- [**Deskripsi Solusi Parallel**](#deskripsi-solusi-parallel)
- [**Analisis Solusi Parallel**](#analisis-solusi-parallel)
- [**Pengukuran Kinerja Solusi Parallel**](#pengukuran-kinerja-solusi-parallel)
- [**Analisis Kinerja Solusi Parallel**](#analisis-kinerja-solusi-parallel)

### Deskripsi Solusi Parallel
Radix sort adalah algoritma sorting yang termasuk dalam kelas non-comparison karena pada dasarnya radix sort hanya melakukan counting sort terhadap tiap digit pada masing-masing elemen input. Disanalah letak tantangan untuk mengembangkan solusi parallel untuk radix sort karena counting sort tidak dapat dikonversi menjadi paralel dengan mudah. Hal ini dikarenakan saat melakukan counting sort, perhitungan setiap iterasinya dependen terhadap perhitungan pada iterasi sebelumnya.

Berangkat dari permasalahan tersebut, kami menggunakan bantuan CUDA untuk melakukan parallel radix sort, dimana kami mencoba untuk memparalelkan beberapa aspek dari subroutine counting sort yang dilakukan dalam radix sort. Bagian yang kami paralelkan adalah pembuatan flag (0 atau 1) untuk tiap elemen array dan proses scatter untuk menyusun ulang posisi elemen-elemen dalam array. Namun, pada solusi kami, masih ada beberapa bagian yang dilakukan oleh host secara serial yaitu proses prefix sum. 

Jumlah thread per block yang kami pakai adalah 128 karena kami mencoba pada server ITB dan hasil parallel radix sortnya lebih cepat. Untuk pemetaan thrad sendiri, program kami akan dibagi menjadi beberapa thread, dimana seperti yang kita tahu bahwa block terdiri dari thread, dan block yang kami pakai sebanyak 128.

### Analisis Solusi Parallel
Kompleksitas dari solusi Parallel kami adalah O(n + 2n/c), sedangkan kompleksitas solusi Serial adalah O(2n + 10). C disini adalah jumlah proses yang dilakukan. Dapat dilihat bahwa semakin banyak process yang tersedia dalam solusi parallel yang kami buat, performanya semakin mendekati 2x lebih cepat dari performa solusi serial. Hal ini dikarenakan kami hanya memparallelkan fungsi pembuatan flags dan scatter saja, sehingga dalam Big O solusi Parallel kami terdapat tambahan sejumlah n, yang merupakan waktu perhitungan prefix sum dimana proses ini masih berjalan secara serial. Sehingga batas peningkatan performa solusi parallel kami adalah 2x.

Solusi ini sebenarnya bisa dibuat lebih efisien apabila proses prefix scan dilakukan secara paralel sepenuhnya misalkan dengan Blelloch Scan.

Pada bagian paralel dari solusi ini, setiap thread melakukan komputasi untuk n/c elemen, misalnya menentukan flag. 

### Pengukuran Kinerja Solusi Parallel
Kami menggunakan metode perbandingan waktu performa untuk mengukur kinerja solusi parallel yang telah kami buat. Kami membandingkan waktu yang dibutuhkan oleh solusi radix sort parallel kami dengan waktu yang dibutuhkan oleh radix sort serial pada array yang isi dan jumlahnya sama persis. Berikut adalah snippet code yang kami pakai untuk mengukur performa solusi kami:

```c
start = clock();

// Move sorting CPU
int* d_arr;
cudaMalloc(&d_arr, sizeof(int) * n);
cudaMemcpy(d_arr, arr, sizeof(int) * n, cudaMemcpyHostToDevice);
radix_sort(d_arr, n);
cudaMemcpy(arr, d_arr, n * sizeof(int), cudaMemcpyDeviceToHost);
cudaFree(d_arr);

end = clock();
cpu_time_used = ((double) (end - start)) * 1000000 / CLOCKS_PER_SEC;
```

Berikut ini adalah hasil pengukuran kinerja solusi kami pada server ITB:

**n = 5000**
```
Serial:
13516020@housting:~/cuda$ ./serial_radix_sort 5000
Sorting in serial...
Serial radix sort finished, time taken: 3476.000000 μs

13516020@housting:~/cuda$ ./serial_radix_sort 5000
Sorting in serial...
Serial radix sort finished, time taken: 3587.000000 μs

13516020@housting:~/cuda$ ./serial_radix_sort 5000
Sorting in serial...
Serial radix sort finished, time taken: 3482.000000 μs

===============================================================================
Parallel:
13516020@housting:~/cuda$ ./radix_sort 5000
Sorting in parallel...
Parallel radix sort finished, time taken: 755029.000000 μs

13516020@housting:~/cuda$ ./radix_sort 5000
Sorting in parallel...
Parallel radix sort finished, time taken: 813429.000000 μs

13516020@housting:~/cuda$ ./radix_sort 5000
Sorting in parallel...
Parallel radix sort finished, time taken: 762473.000000 μs

```

**n = 50000**
```
Serial:
13516020@housting:~/cuda$ ./serial_radix_sort 50000
Sorting in serial...
Serial radix sort finished, time taken: 24206.000000 μs

13516020@housting:~/cuda$ ./serial_radix_sort 50000
Sorting in serial...
Serial radix sort finished, time taken: 30431.000000 μs

13516020@housting:~/cuda$ ./serial_radix_sort 50000
Sorting in serial...
Serial radix sort finished, time taken: 24103.000000 μs
===============================================================================
Parallel:
13516020@housting:~/cuda$ ./radix_sort 50000
Sorting in parallel...
Parallel radix sort finished, time taken: 635070.000000 μs

13516020@housting:~/cuda$ ./radix_sort 50000
Sorting in parallel...
Parallel radix sort finished, time taken: 620376.000000 μs

13516020@housting:~/cuda$ ./radix_sort 50000
Sorting in parallel...
Parallel radix sort finished, time taken: 629186.000000 μs
```

**n = 100000**
```
Serial:
13516020@housting:~/cuda$ ./serial_radix_sort 100000
Sorting in serial...
Serial radix sort finished, time taken: 40066.000000 μs

13516020@housting:~/cuda$ ./serial_radix_sort 100000
Sorting in serial...
Serial radix sort finished, time taken: 39948.000000 μs

13516020@housting:~/cuda$ ./serial_radix_sort 100000
Sorting in serial...
Serial radix sort finished, time taken: 39953.000000 μs
===============================================================================
Parallel:
13516020@housting:~/cuda$ ./radix_sort 100000
Sorting in parallel...
Parallel radix sort finished, time taken: 659299.000000 μs

13516020@housting:~/cuda$ ./radix_sort 100000
Sorting in parallel...
Parallel radix sort finished, time taken: 685127.000000 μs

13516020@housting:~/cuda$ ./radix_sort 100000
Sorting in parallel...
Parallel radix sort finished, time taken: 622940.000000 μs
```

**n = 200000**
```
Serial:
13516020@housting:~/cuda$ ./serial_radix_sort 200000
Sorting in serial...
Serial radix sort finished, time taken: 83413.000000 μs

13516020@housting:~/cuda$ ./serial_radix_sort 200000
Sorting in serial...
Serial radix sort finished, time taken: 73147.000000 μs

13516020@housting:~/cuda$ ./serial_radix_sort 200000
Sorting in serial...
Serial radix sort finished, time taken: 82367.000000 μs
===============================================================================
Parallel:
13516020@housting:~/cuda$ ./radix_sort 200000
Sorting in parallel...
Parallel radix sort finished, time taken: 760482.000000 μs

13516020@housting:~/cuda$ ./radix_sort 200000
Sorting in parallel...
Parallel radix sort finished, time taken: 760531.000000 μs

13516020@housting:~/cuda$ ./radix_sort 200000
Sorting in parallel...
Parallel radix sort finished, time taken: 761216.000000 μs
```

**n = 400000**
```
Serial:
13516020@housting:~/cuda$ ./serial_radix_sort 400000
Sorting in serial...
Serial radix sort finished, time taken: 142686.000000 μs

13516020@housting:~/cuda$ ./serial_radix_sort 400000
Sorting in serial...
Serial radix sort finished, time taken: 143178.000000 μs

13516020@housting:~/cuda$ ./serial_radix_sort 400000
Sorting in serial...
Serial radix sort finished, time taken: 141915.000000 μs
===============================================================================
Parallel:
13516020@housting:~/cuda$ ./radix_sort 400000
Sorting in parallel...
Parallel radix sort finished, time taken: 785813.000000 μs

13516020@housting:~/cuda$ ./radix_sort 400000
Sorting in parallel...
Parallel radix sort finished, time taken: 786641.000000 μs

13516020@housting:~/cuda$ ./radix_sort 400000
Sorting in parallel...
Parallel radix sort finished, time taken: 811905.000000 μs
```

### Analisis Kinerja Solusi Parallel
Memang pada pengujian pada jumlah N yang diberikan, solusi parallel kami tidak ada yang lebih cepat dari serial. Hanya saja, jika kita lihat trend dari waktu pengujian, waktu yang dibutuhkan parallel radix sort selalu sama, berkisar pada angka 600000 - 700000 μs, sedangkan waktu yang dibutuhkan serial radix sort linear, naik seiring dengan jumlah N. Dapat disimpulkan bahwa untuk nilai N yang besar, maka solusi parallel akan jauh lebih cepat dari solusi serial radix sort.

## Author
### Faza Fahleraz
```
Email: 13516095@std.stei.itb.ac.id
Github: https://github.com/ffahleraz
Task:
* Membuat counting sort parallel
* Membuat scatter parallel
* Membuat serial radix sort
* Membuat Dokumentasi
```

### Nicholas Rianto Putra
```
Email: 13516020@std.stei.itb.ac.id
Github: https://github.com/nicholaz99
Task:
* Membuat Main Program
* Membuat Program mengoutputkan hasil ke dalam file .txt pada folder output
* Membuat prefix sum
* Membuat Dokumentasi
```
